import datetime
import uuid
from unittest import mock

import pytest
from django.utils.timezone import now
from pyquery import PyQuery

from lingo.basket.models import Basket, BasketLine, BasketLineItem
from lingo.invoicing.models import Credit, DraftInvoice, DraftInvoiceLine, Invoice, Regie
from tests.utils import login

pytestmark = pytest.mark.django_db


def test_basket_detail(app, simple_user):
    resp = app.get('/basket/')
    assert resp.location.endswith('/login/?next=/basket/')
    app = login(app, username='user', password='user')

    # no basket object
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' not in resp
    assert '/basket/validate/' not in resp
    assert '/basket/cancel/' not in resp

    # basket without lines
    regie = Regie.objects.create(label='Foo')
    invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    basket = Basket.objects.create(regie=regie, draft_invoice=invoice, payer_nameid='ab' * 16)
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' not in resp
    assert '/basket/validate/' not in resp
    assert '/basket/cancel/' not in resp

    # a not closed line
    line = BasketLine.objects.create(
        basket=basket,
        closed=False,
        user_first_name='First1',
        user_last_name='Last1',
    )
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' not in resp
    assert '/basket/validate/' not in resp
    assert '/basket/cancel/' not in resp

    # line is closed but empty
    line.closed = True
    line.save()
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' in resp
    assert '/basket/validate/' in resp
    assert '/basket/cancel/' in resp

    # add some items, group_items is False
    BasketLineItem.objects.create(
        line=line,
        label='Repas',
        subject='Réservation',
        details='Lun 06/11, Mar 07/11',
        quantity=2,
        unit_amount=3,
    )
    BasketLineItem.objects.create(
        line=line,
        label='Repas',
        subject='Réservation',
        details='Jeu 09/11',
        quantity=1,
        unit_amount=3,
    )
    BasketLineItem.objects.create(
        line=line,
        label='Repas',
        subject='Annulation',
        details='Ven 10/11',
        quantity=-1,
        unit_amount=3,
    )
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 3
    assert [PyQuery(li).text() for li in resp.pyquery('ul.basket li')] == [
        'First1 Last1 - Repas - Annulation Ven 10/11 -3.00€',
        'First1 Last1 - Repas - Réservation Jeu 09/11 3.00€',
        'First1 Last1 - Repas - Réservation Lun 06/11, Mar 07/11 6.00€',
    ]
    assert '/basket/invoice/pdf/' in resp
    assert '/basket/validate/' in resp
    assert '/basket/cancel/' in resp

    # group items
    line.group_items = True
    line.save()
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 2
    assert [PyQuery(li).text() for li in resp.pyquery('ul.basket li')] == [
        'First1 Last1 - Repas - Annulation Ven 10/11 -3.00€',
        'First1 Last1 - Repas - Réservation Lun 06/11, Mar 07/11, Jeu 09/11 9.00€',
    ]
    assert '/basket/invoice/pdf/' in resp
    assert '/basket/validate/' in resp
    assert '/basket/cancel/' in resp

    # not closed line
    line.closed = False
    line.save()
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' not in resp
    assert '/basket/validate/' not in resp
    assert '/basket/cancel/' not in resp

    # basket payer_nameid is wrong
    line.closed = True
    line.save()
    basket.payer_nameid = uuid.uuid4()
    basket.save()
    resp = app.get('/basket/')
    assert 'My basket' in resp
    assert len(resp.pyquery('ul.basket li')) == 0
    assert '/basket/invoice/pdf/' not in resp
    assert '/basket/validate/' not in resp
    assert '/basket/cancel/' not in resp

    # check status
    basket.payer_nameid = 'ab' * 16
    basket.save()
    for status in ['open', 'tobepaid']:
        basket.status = status
        basket.save()
        resp = app.get('/basket/')
        assert 'My basket' in resp
        assert len(resp.pyquery('ul.basket li')) == 2
        assert '/basket/invoice/pdf/' in resp
        if status == 'open':
            assert '/basket/validate/' in resp
        else:
            assert '/basket/validate/' not in resp
        assert '/basket/cancel/' in resp
    for status in ['cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        resp = app.get('/basket/')
        assert 'My basket' in resp
        assert len(resp.pyquery('ul.basket li')) == 0
        assert '/basket/invoice/pdf/' not in resp
        assert '/basket/validate/' not in resp
        assert '/basket/cancel/' not in resp


def test_basket_invoice_pdf(app, simple_user):
    resp = app.get('/basket/invoice/pdf/')
    assert resp.location.endswith('/login/?next=/basket/invoice/pdf/')
    app = login(app, username='user', password='user')

    # no basket object
    app.get('/basket/invoice/pdf/', status=404)

    # basket
    regie = Regie.objects.create(label='Foo')
    invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    basket = Basket.objects.create(regie=regie, draft_invoice=invoice, payer_nameid='ab' * 16)
    app.get('/basket/invoice/pdf/', status=200)

    # basket payer_nameid is wrong
    basket.payer_nameid = uuid.uuid4()
    basket.save()
    app.get('/basket/invoice/pdf/', status=404)

    # check status
    basket.invoice = basket.draft_invoice.promote()
    basket.payer_nameid = 'ab' * 16
    basket.save()
    for status in ['open', 'tobepaid']:
        basket.status = status
        basket.save()
        resp = app.get('/basket/invoice/pdf/?html', status=200)
        if status == 'open':
            assert 'TEMPORARY-%s' % basket.draft_invoice.pk in resp
        else:
            assert 'F%02d-%s-0000001' % (regie.pk, basket.invoice.created_at.strftime('%y-%m')) in resp
        assert 'Total amount to be paid before' not in resp
    for status in ['cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        app.get('/basket/invoice/pdf/', status=404)


def test_basket_validate(app, simple_user):
    resp = app.get('/basket/validate/')
    assert resp.location.endswith('/login/?next=/basket/validate/')
    app = login(app, username='user', password='user')

    # no basket object
    app.get('/basket/validate/', status=404)

    # basket without line
    regie = Regie.objects.create(label='Foo')
    invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=invoice,
        payer_nameid='ab' * 16,
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    app.get('/basket/validate/', status=404)

    # a not closed line
    line = BasketLine.objects.create(
        basket=basket,
        closed=False,
        user_external_id='user:1',
    )
    app.get('/basket/validate/', status=404)

    # line is closed, but wrong payer_nameid
    line.closed = True
    line.save()
    basket.payer_nameid = uuid.uuid4()
    basket.save()
    app.get('/basket/validate/', status=404)

    # good payer_nameid
    basket.payer_nameid = 'ab' * 16
    basket.save()
    resp = app.get('/basket/validate/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    assert resp.location.endswith('/basket/')
    basket.refresh_from_db()
    assert basket.status == 'tobepaid'
    assert basket.validated_at is not None
    assert Invoice.objects.count() == 1
    invoice = Invoice.objects.latest('pk')
    assert basket.invoice == invoice
    assert basket.credit is None

    # wrong status
    for status in ['tobepaid', 'cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        app.get('/basket/validate/', status=404)

    # check callback
    basket.status = 'open'
    basket.save()
    line.validation_callback_url = 'http://validation1.com'
    line.credit_callback_url = 'http://validation1.com'
    line.save()
    BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:2',
        validation_callback_url='http://validation2.com',
        credit_callback_url='http://validation2.com',
    )
    resp = app.get('/basket/validate/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == {
        'http://validation1.com/',
        'http://validation2.com/',
    }

    # basket is expired
    basket.expiry_at = now()
    basket.status = 'open'
    basket.save()
    app.get('/basket/validate/', status=404)


def test_basket_validate_generate_invoice(app, simple_user):
    app = login(app, username='user', password='user')

    regie = Regie.objects.create(label='Foo')
    draft_invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    DraftInvoiceLine.objects.create(
        slug='event-a-foo-bar',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=draft_invoice,
        quantity=1,
        unit_amount=1,
        user_external_id='user:1',
    )
    # invoice total amount is positive
    draft_invoice.refresh_from_db()
    assert draft_invoice.total_amount == 1
    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=draft_invoice,
        payer_nameid='ab' * 16,
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:1',
        group_items=False,
    )

    resp = app.get('/basket/validate/')
    resp = resp.form.submit()
    assert resp.location.endswith('/basket/')
    basket.refresh_from_db()
    assert basket.status == 'tobepaid'
    assert basket.validated_at is not None
    assert basket.completed_at is None
    invoice = Invoice.objects.latest('pk')
    assert basket.invoice == invoice
    assert invoice.total_amount == 1
    assert Credit.objects.count() == 0

    # total is zero
    DraftInvoiceLine.objects.create(
        slug='event-b-foo-bar',
        label='Event B',
        event_date=datetime.date(2022, 9, 1),
        invoice=draft_invoice,
        quantity=-1,
        unit_amount=1,
        user_external_id='user:1',
    )
    draft_invoice.refresh_from_db()
    assert draft_invoice.total_amount == 0
    basket.status = 'open'
    basket.save()

    resp = app.get('/basket/validate/')
    resp = resp.form.submit()
    assert resp.location.endswith('/basket/')
    basket.refresh_from_db()
    assert basket.status == 'tobepaid'
    assert basket.validated_at is not None
    assert basket.completed_at is None
    invoice = Invoice.objects.latest('pk')
    assert basket.invoice == invoice
    assert invoice.total_amount == 0
    assert Credit.objects.count() == 0


def test_basket_validate_generate_credit(app, simple_user):
    app = login(app, username='user', password='user')

    regie = Regie.objects.create(label='Foo')
    draft_invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
        payer_external_id='payer:1',
        payer_first_name='First',
        payer_last_name='Last',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
    )
    DraftInvoiceLine.objects.create(
        slug='event-a-foo-bar',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=draft_invoice,
        quantity=-1,
        unit_amount=1,
        user_external_id='user:1',
        user_first_name='First1',
        user_last_name='Last1',
    )
    draft_invoice.refresh_from_db()
    assert draft_invoice.total_amount == -1
    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=draft_invoice,
        payer_nameid='ab' * 16,
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    line = BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:1',
    )

    resp = app.get('/basket/validate/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    basket.refresh_from_db()
    assert basket.status == 'completed'
    assert basket.validated_at is not None
    assert basket.completed_at is not None
    assert basket.invoice is None
    credit = Credit.objects.latest('pk')
    assert basket.credit == credit
    assert credit.label == 'Credit from %s' % datetime.date.today().strftime('%d/%m/%Y')
    assert credit.total_amount == 1
    assert credit.regie == regie
    assert credit.payer_external_id == 'payer:1'
    assert credit.payer_first_name == 'First'
    assert credit.payer_last_name == 'Last'
    assert credit.payer_address == '41 rue des kangourous\n99999 Kangourou Ville'
    assert credit.lines.count() == 1
    (line1,) = credit.lines.all().order_by('pk')
    assert line1.event_date == datetime.date(2022, 9, 1)
    assert line1.slug == 'event-a-foo-bar'
    assert line1.label == 'Event A'
    assert line1.quantity == 1
    assert line1.unit_amount == 1
    assert line1.total_amount == 1
    assert line1.user_external_id == 'user:1'
    assert line1.user_first_name == 'First1'
    assert line1.user_last_name == 'Last1'
    assert Invoice.objects.count() == 0

    # check callback
    basket.status = 'open'
    basket.save()
    line.validation_callback_url = 'http://validation1.com'
    line.credit_callback_url = 'http://credit1.com'
    line.save()
    BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:2',
        validation_callback_url='http://validation2.com',
        credit_callback_url='http://credit2.com',
    )
    resp = app.get('/basket/validate/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    basket.refresh_from_db()
    credit = Credit.objects.latest('pk')
    assert basket.credit == credit
    assert Invoice.objects.count() == 0
    assert {x[0][0].url for x in mock_send.call_args_list} == {
        'http://validation1.com/',
        'http://validation2.com/',
        'http://credit1.com/',
        'http://credit2.com/',
    }


def test_basket_cancel(app, simple_user):
    resp = app.get('/basket/cancel/')
    assert resp.location.endswith('/login/?next=/basket/cancel/')
    app = login(app, username='user', password='user')

    # no basket object
    app.get('/basket/cancel/', status=404)

    # basket without line
    regie = Regie.objects.create(label='Foo')
    draft_invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    invoice = Invoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=draft_invoice,
        invoice=invoice,
        payer_nameid='ab' * 16,
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    app.get('/basket/cancel/', status=404)

    # a not closed line
    line = BasketLine.objects.create(
        basket=basket,
        closed=False,
        user_external_id='user:1',
    )
    app.get('/basket/cancel/', status=404)

    # line is closed, but wrong payer_nameid
    line.closed = True
    line.save()
    basket.payer_nameid = uuid.uuid4()
    basket.save()
    app.get('/basket/cancel/', status=404)

    # good payer_nameid
    basket.payer_nameid = 'ab' * 16
    basket.save()
    resp = app.get('/basket/cancel/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    assert resp.location.endswith('/basket/')
    basket.refresh_from_db()
    assert basket.status == 'cancelled'
    assert basket.cancelled_at is not None
    invoice.refresh_from_db()
    assert invoice.cancelled_at is not None

    basket.status = 'tobepaid'
    basket.cancelled_at = None
    basket.save()
    resp = app.get('/basket/cancel/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    assert resp.location.endswith('/basket/')
    basket.refresh_from_db()
    assert basket.status == 'cancelled'
    assert basket.cancelled_at is not None

    # wrong status
    for status in ['cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        app.get('/basket/cancel/', status=404)

    # check callback
    basket.status = 'open'
    basket.save()
    line.cancel_callback_url = 'http://cancellation1.com'
    line.save()
    BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:2',
        cancel_callback_url='http://cancellation2.com',
    )
    resp = app.get('/basket/cancel/')
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = resp.form.submit()
    assert {x[0][0].url for x in mock_send.call_args_list} == {
        'http://cancellation1.com/',
        'http://cancellation2.com/',
    }

    # basket is expired
    basket.expiry_at = now()
    basket.status = 'open'
    basket.save()
    app.get('/basket/cancel/')
