import datetime
from unittest import mock

import pytest
from django.utils.timezone import now

from lingo.basket.models import Basket, BasketLine
from lingo.invoicing.models import DraftInvoice, Invoice, Regie

pytestmark = pytest.mark.django_db


def test_basket_expiration():
    regie = Regie.objects.create(label='Foo')
    draft_invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    invoice = Invoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=draft_invoice,
        invoice=invoice,
        expiry_at=now() + datetime.timedelta(minutes=1),
    )
    line = BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:1',
    )

    for status in ['open', 'tobepaid', 'cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        # expiry_at is not passed, no changes
        Basket.expire_baskets()
        basket.refresh_from_db()
        assert basket.status == status
        assert basket.expired_at is None

    # open basket, expire it immediatly after expiry_at is passed
    basket.status = 'open'
    basket.expiry_at = now()
    basket.save()
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        Basket.expire_baskets()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    basket.refresh_from_db()
    assert basket.status == 'expired'
    assert basket.expired_at is not None
    invoice.refresh_from_db()
    assert invoice.cancelled_at is not None

    # tobepaid basket, expire it 1 hour after expiry_at is passed
    line.expiration_callback_url = 'http://expiration1.com'
    line.save()
    BasketLine.objects.create(
        basket=basket,
        closed=True,
        user_external_id='user:2',
        expiration_callback_url='http://expiration2.com',
    )

    basket.expired_at = None
    basket.status = 'tobepaid'
    basket.expiry_at = now()
    basket.save()
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        Basket.expire_baskets()
    assert {x[0][0].url for x in mock_send.call_args_list} == set()
    basket.refresh_from_db()
    assert basket.status == 'tobepaid'
    assert basket.expired_at is None

    basket.expiry_at = now() - datetime.timedelta(minutes=60)
    basket.save()
    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        Basket.expire_baskets()
    assert {x[0][0].url for x in mock_send.call_args_list} == set(
        {'http://expiration1.com/', 'http://expiration2.com/'}
    )
    basket.refresh_from_db()
    assert basket.status == 'expired'
    assert basket.expired_at is not None

    # other status, no changes
    basket.expired_at = None
    basket.expiry_at = now()
    for status in ['cancelled', 'expired', 'completed']:
        basket.status = status
        basket.save()
        Basket.expire_baskets()
        basket.refresh_from_db()
        assert basket.status == status
        assert basket.expired_at is None
